package statej;

import java.io.*;

import statej.rewriter.JavaWriter;
import statej.typeChecker.SemanticChecker;
import statej.typeChecker.SymbolTable;
import statej.typeChecker.SymbolTableConstructor;


import statej.ast.Node;

import java_cup.runtime.Symbol;
import statej.parser.Lexer;
import statej.parser.LexingException;
import statej.parser.parser;

public class sjc
{
	private static void printUsage()
	{
		System.out.println("Usage: java sjc filename*");
		System.out.println("    for at least one filename");
		System.exit(0);
	}
	
	public static void main(String[] args)
	{
		if (args.length == 0)
		{
			printUsage();
		}

		for (String param : args)
		{
			compile(param);
		}
		
		System.exit(0);
	}
	
	private static void compile(String filename)
	{
        PrintWriter err = new PrintWriter(System.out, true);
        Reader input = makeReader(filename);

        //make an ast
        parser parser = new parser(new Lexer(input, err), err);
        Symbol result = null;
        try
        {
            result = parser.parse();
        }
        catch (LexingException e)
        {
            err.println("Parsing failed during lexing.");
        }
        catch (Exception e)
        {
            //An exception occured during parsing, however we will know this
            //anyway from the error count so do nothing
        }

        if (parser.getErrorCount() != 0)
        {
            err.println("Parsing failed: " + parser.getErrorCount() + " error(s).");
            close(input, null);
            return;
        }
		
		//scope and type check
        Node root = (Node)result.value;
        if (!checkSemantics(root, err))
        {
            close(input, null);
            return;
        }
        	
		
		//write java file
        Writer output = makeWriter(makeOutputFilename(filename));
        JavaWriter jw = new JavaWriter(err);
        String outString = (String) root.accept(jw);
        try
		{
			output.write(outString);
			err.println("Java written to " + makeOutputFilename(filename));
		}
		catch (IOException e)
		{
			err.println("Error writing output: " + e.getMessage());
		}
        
        
        close(input, output);
	}
	
    //check the semantics of the given program
    private static boolean checkSemantics(Node root, PrintWriter err)
    {
        //1st pass: build the symbol table
        SymbolTableConstructor stc = new SymbolTableConstructor(err);
        root.accept(stc);
        int errors = stc.getErrorCount();
        int warnings = stc.getWarningCount();
        if (errors > 0)
        {
            err.println("Semantic analysis failed with " + errors + " error(s) during symbol table construction");
            return false;
        }
        SymbolTable symbolTable = stc.getSymbolTable();

        //all classes are now in the super table, resolve any super class references
        if (!symbolTable.resolveSuperClasses(err))
        {
            err.println("Semantic analysis failed during super class resolution");
            return false;
        }


        //2nd pass: check types and scopes
        SemanticChecker  tsc = new SemanticChecker(symbolTable, err);
        root.accept(tsc);
        errors = tsc.getErrorCount();
        warnings += tsc.getWarningCount();
        if (errors > 0)
        {
            err.println("Semantic analysis failed with " + errors + " error(s) during semantics checking");
            return false;
        }

        err.println("Semantic analysis complete - 0 errors, " + warnings + " warnings.");

        return true;
    }
	
	private static  void close(Reader input, Writer output)
	{
        //close the files
        try
		{
			input.close();
		}
		catch (IOException e)
		{
			// do nothing
		}
		if (output != null)
		{
	        try
			{
				output.close();
			}
			catch (IOException e)
			{
				// do nothing
			}
		}
	}

	private static Reader makeReader(String filename)
	{
		try
		{
			return new BufferedReader(new FileReader(filename));
		}
		catch (FileNotFoundException e)
		{
			System.out.println("Error opening file: " + filename);
			System.exit(0);
			return null;
		}
	}
	
	private static Writer makeWriter(String filename)
	{
		try
		{
			return new BufferedWriter(new FileWriter(filename));
		}
		catch (Exception e)
		{
			System.out.println("Error opening file: " + filename);
			System.exit(0);
			return null;
		}
	}
	
	private static String makeOutputFilename(String filename)
	{
        String outFilename;
        int dotPos = filename.indexOf('.');
        if (dotPos < 0)
            outFilename = filename + ".java";
        else
            outFilename = filename.substring(0, dotPos) + ".java";
        
        return outFilename;
	}
}
