package statej.rewriter;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import statej.Visitor;
import statej.ast.*;
import statej.typeChecker.ClassSymbol;
import statej.typeChecker.MethodSymbol;
import statej.typeChecker.StateSymbol;

public class JavaWriter extends Visitor
{
	//a list of all methods declared outside of any state
	//protected List<MethodSymbol> defMethods = new ArrayList<MethodSymbol>();
	
	public JavaWriter(PrintWriter err)
	{
		super(err);
	}
	
	public Object visit(AddNode node)
	{
		return node.visitLeft(this) + " + " + node.visitRight(this);
	}

	public Object visit(AndNode node)
	{
		return node.visitLeft(this) + " & " + node.visitRight(this);
	}

	public Object visit(ArgNode node)
	{
		return node.toString();
	}

	public Object visit(AssNode node)
	{
		return node.visitTarget(this) + " = " + node.visitValue(this);
	}

	public Object visit(BooleanNode node)
	{
		return node.toString();
	}

	public Object visit(ClassNode node)
	{
        String result = "class " + node.getName();
        String superClass = node.getSuperClass();
        if (superClass == null || superClass.equals("Object"))
        {
            result += " extends statej.runtime.StateObject";
        }
        else
        {
            result += " extends " + superClass;
        }
        
        //ClassNode superClassNode = (ClassNode)node.getSymbol().getSymbolTable().get(superClass);
        //if (superClassNode != null && superClassNode.)
        
        String slots = "";

        //write the fields and name mangled methods
        for (Object slot : node.visitSlots(this))
        {
        	slots += slot + "\n";
        }

        
        //write the co-ordinating state methods
        Map<String, List<StateSymbol>> stateMethods = node.getSymbol().getStateMethods();
        Set<String> keys = stateMethods.keySet();
        for (String key : keys)
        {
        	List<StateSymbol> states = stateMethods.get(key);
        	slots = buildMethods(key, slots, node, states);
        	
        	//remove any methods that get a coordinating method
        	//for (MethodSymbol m : defMethods)
        	//{
        	//	if (key.equals(m.getKey()))
        	//		defMethods.remove(m);
        	//}
        }
        
        //write the default only methods
    	//for (MethodSymbol m : defMethods)
    	//{
        //	slots = buildMethods(m.getKey(), slots, node, new ArrayList<StateSymbol>());
    	//}
        
		return result + "\n{\n" + slots + "\n}\n";
	}
	
	//produces the coordinating methods that call the methods for a certain state
	private String buildMethods(String key, String slots, ClassNode node, List<StateSymbol> states)
	{
    	MethodSymbol meth;
    	if (states.size() == 0)
        	meth = (MethodSymbol) node.getSymbol().getSymbolTable().get(key);
    	else
    		meth = (MethodSymbol) node.getSymbol().getSymbolTable().get(states.get(0).getName() + "$$" + key);
    	MethodNode methNode = (MethodNode)meth.getNode();
    	
    	boolean isVoid = methNode.getType().equals("void");
		String name = methNode.getName();
		Object[] params = methNode.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);

        String pairType;
        if (isVoid)
        	pairType = "Object";
        else
        	pairType = methNode.getType();

       	String actuals = "";
    	for (Object p : methNode.getParamNames())
    	{
    		actuals += p + ", ";
    	}
        if (actuals.length() >= 2)
        	actuals = actuals.substring(0, actuals.length() - 2);
        
        
        slots += methNode.getAccess() + " " + methNode.getType() + " " + name + "(" + paramString + ")\n{\n";
        
        slots += "synchronized(this.__lock)\n{\nwhile (true)\n{\n";
        slots += "statej.runtime.Pair<Boolean, " + pairType + "> __result = __" + name +  "(" + actuals + ");\n";
        
        slots += "if (__result.getX())\nreturn";
        if (isVoid)
        	slots += ";\n";
        else
        	slots += " __result.getY();\n";
        slots += "try\n{\nthis.__lock.wait();\n}\ncatch (InterruptedException __ie) {}\n";
        slots += "}\n}\n}\n";

        String access = methNode.getAccess().toString();
        if (!access.equals("private"))
        	access = "protected";
        slots += access + " statej.runtime.Pair<Boolean, " + pairType + "> __" + name + "(" + paramString + ")\n{\nswitch(this.__curState)\n{\n";
        
        for (StateSymbol state : states)
        {
        	MethodSymbol curMethSym = (MethodSymbol) node.getSymbol().getSymbolTable().get(state.getName() + "$$" + key);
        	MethodNode curMethNode = (MethodNode)curMethSym.getNode();
        	slots += "case " + state.getStateNumber() + ":\n";
        	actuals = "";
        	for (Object p : curMethNode.getParamNames())
        	{
        		actuals += p + ", ";
        	}
            if (actuals.length() >= 2)
            	actuals = actuals.substring(0, actuals.length() - 2);
        	
            int stateNumber = state.getStateNumber();
            if (curMethNode.getAsState() != null)
            {
            	stateNumber = ((StateSymbol)node.getSymbol().getSymbolTable().get(StateSymbol.KEY_PREFIX + curMethNode.getAsState())).getStateNumber();
            }
            
            if (isVoid)
            	slots += "this.__" + stateNumber + "__" + curMethSym.getName() + "(" + actuals + ");\nreturn new statej.runtime.Pair<Boolean, Object>(true, null);\n";
            else
            	slots += "return new statej.runtime.Pair<Boolean, " + pairType + ">(true, this.__" + stateNumber + "__" + curMethSym.getName() + "(" + actuals + "));\n";
        }
        
        String superName = node.getSuperClass();
        boolean hasSuperMethod = true;
        if (superName == null)
        	hasSuperMethod = false;
        else
        {
        	ClassSymbol superSymbol = (ClassSymbol) node.getSymbol().getSymbolTable().get(superName);
        	if (superSymbol == null)
        		hasSuperMethod = false;
        	else
        	{
        		hasSuperMethod = superSymbol.hasStateMethod(meth.getKey(null));
        	}
        }
        if (hasSuperMethod) //if a method exists in the superclass we need to dispatch there
        {
        	actuals = "";
        	for (Object p : methNode.getParamNames())
        	{
        		actuals += p + ", ";
        	}
            if (actuals.length() >= 2)
            	actuals = actuals.substring(0, actuals.length() - 2);

            if (isVoid)
               	slots += "default:\nsuper." + meth.getName() + "(" + actuals + ");\nreturn new statej.runtime.Pair<Boolean, Object>(true, null);\n}\n}\n";
            else
            	slots += "default:\nreturn new statej.runtime.Pair<Boolean, " + pairType + ">(true, super." + meth.getName() + "(" + actuals + "));\n}\n}\n";
        }
        else
        	slots += "default:\nreturn new statej.runtime.Pair<Boolean, " + pairType + ">(false, null);\n}\n}\n";
        
        return slots;
	}

	public Object visit(CompoundNode node)
	{
		Object[] stmts = node.visitChildren(this);
		String[] eols = node.getEols();
		String result = "";
		for (int i = 0; i < stmts.length; ++i)
		{
			result += stmts[i] + eols[i] + "\n";
		}
		
		Object synch = node.visitSynch(this);
		if (synch == null)
			return "{\n" + result + "}\n";
		else
			return "synchronized (" + synch + ")\n{\n" + result + "}\n";
	}


    public Object visit(BracketNode node)
    {
        return "(" + node.visitChild(this) + ")";
    }

    public Object visit(StateChangeNode node)
	{
    	if (node.getName().equals("previous"))
    		return "__changeState(__prevState);\n";
    	else
    		return "__changeState(" + node.getStateSymbol().getStateNumber() + ");\n";
	}

    public Object visit(StateDefNode node)
	{
    	Object[] meths = node.visitMethods(this);
    	String result = "";
    	for (Object s : meths)
    		result += s + "\n";
		return result;
	}

	public Object visit(ConstructorNode node)
	{
		Object[] params = node.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);

        return node.getAccess() + " " + node.getName() + "(" + paramString + ")\n" + node.visitBody(this);
	}

	public Object visit(DeclNode node)
	{
		return node.getType() + " " + node.getName() + " = " + node.visitValue(this);
	}

	public Object visit(DivNode node)
	{
		return node.visitLeft(this) + " / " + node.visitRight(this);
	}

	public Object visit(ComparNode node)
	{
		return node.visitLeft(this) + " " + node.getOp() + " " + node.visitRight(this);
	}

	public Object visit(SpawnNode node)
	{
		return "{\nThread __t = new Thread() {\npublic void run()\n{\n" + node.visitStmt(this) + "\n}\n};\n__t.start();}\n";
	}

	public Object visit(FieldAccessNode node)
	{
		return node.visitTarget(this) + "." + node.getName();
	}

	public Object visit(FieldNode node)
	{
		return node.getAccess() + " " + node.getType() + " " + node.getName() + " = " + node.visitValue(this) + ";";
	}

	public Object visit(ForNode node)
	{
		return "for (" + node.visitInitExpr(this) + "; " + node.visitTestExpr(this) + "; " + node.visitIncExpr(this) + ")\n" + node.visitLoopStmt(this) + "\n";
	}

	public Object visit(IfNode node)
	{
		return "if (" + node.visitTest(this) + ")\n" + node.visitThen(this) + "\n";
	}

	public Object visit(ListNode node)
	{
		Node[] elements = node.toArray();
		StringBuilder result = new StringBuilder();
		for (Node n : elements)
		{
			result.append(n.accept(this));
			result.append('\n');
		}
		
		return result.toString();
	}

	public Object visit(MethodInvocNode node)
	{
		Object[] params = node.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);
        
		return node.visitTarget(this) + "." + node.getName() + "(" + paramString + ")";
	}

	public Object visit(MethodNode node)
	{
		String name = node.getName();
		String access = node.getAccess().toString();
		if (node.getSymbol().getState() != null)
		{
			name = "__" + node.getSymbol().getState().getStateNumber() + "__" + name;
			if (!access.equals("private"))
				access = "protected";
		}
		//else
		//{
		//	name = "__0__" + name;
		//	if (!access.equals("private"))
		//		access = "protected";
		//	
		//	defMethods.add(node.getSymbol());
		//}
		Object[] params = node.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);

        String synchString = node.isSynchonized() ? " synchronized" : "";
        
        if (node.getAsState() != null)
        	return "//method called using as:\n//" + access + synchString + " " + node.getType() + " " + name + "(" + paramString + ")\n";
        else
        	return access + synchString + " " + node.getType() + " " + name + "(" + paramString + ")\n" + node.visitBody(this);
	}

	public Object visit(MultNode node)
	{
		return node.visitLeft(this) + " * " + node.visitRight(this);
	}

	public Object visit(NegNode node)
	{
		return "-" + node.visitChild(this);
	}

	public Object visit(NewNode node)
	{
		Object[] params = node.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);
        
		return "new " + node.getClassName() + "(" + paramString + ")";
	}

	public Object visit(NotNode node)
	{
		return "!" + node.visitChild(this);
	}

	public Object visit(NullNode node)
	{
		return "null";
	}

	public Object visit(NumberNode node)
	{
		return node.toString();
	}

	public Object visit(OrNode node)
	{
		return node.visitLeft(this) + " | " + node.visitRight(this);
	}

	public Object visit(ImportNode node)
	{
		return "import " + node.getImportString() + ";";
	}

	public Object visit(ReturnNode node)
	{
		return "return " + node.visitValue(this);
	}

	public Object visit(ProgramNode node)
	{
		String result = "";
		Object pack = node.visitPackage(this);
		if (pack != null)
			result += pack + "\n\n"; 
		
		Object[] imports = node.visitImports(this);
		for (Object s : imports)
			result += s + "\n";
		
		result += "\n";
		
		Object[] classes = node.visitClasses(this);
		for (Object s : classes)
			result += s + "\n";
		
		return  result;
	}

	public Object visit(StringNode node)
	{
		return node.toString();
	}

	public Object visit(SubNode node)
	{
		return node.visitLeft(this) + " - " + node.visitRight(this);
	}

	public Object visit(PackageNode node)
	{
		return node.toString();
	}

	public Object visit(SuperCallNode node)
	{
		Object[] params = node.visitParams(this);
        String paramString =  "";
        for (int i = 0; i < params.length; ++i)
        {
        	paramString += params[i].toString() + ", ";
        }
        if (paramString.length() >= 2)
        	paramString = paramString.substring(0, paramString.length() - 2);

		
		return "super(" + paramString + ")";
	}

	public Object visit(VarNode node)
	{
		return node.toString();
	}

}
